@@ -2,7 +2,7 @@
  *      ipvsadm - IP Virtual Server ADMinistration program
  *                for IPVS NetFilter Module in kernel 2.4
  *
- *      Version: $Id: ipvsadm.c,v 1.18 2001/11/23 14:34:15 wensong Exp $
+ *      Version: $Id: ipvsadm.c,v 1.19.2.2 2003/04/11 14:02:38 wensong Exp $
  *
  *      Authors: Wensong Zhang <wensong@linuxvirtualserver.org>
  *               Peter Kese <peter.kese@ijs.si>
@@ -108,12 +108,6 @@
 #include <netinet/udp.h>
 #include <netinet/tcp.h>
 
-#ifdef HAVE_NET_IP_VS_H
-#include <net/ip_vs.h>
-#else
-#include "ip_vs.h"
-#endif
-
 #ifdef HAVE_POPT
 #include "popt.h"
 #define IPVS_OPTION_PROCESSING          "popt"
@@ -124,13 +118,13 @@
 #include "config_stream.h"
 #include "libipvs/libipvs.h"
 
-#define IPVSADM_VERSION_NO              "v" VERSION
-#define IPVSADM_VERSION_DATE            "2001/11/04"
-#define IPVSADM_VERSION         IPVSADM_VERSION_NO " " IPVSADM_VERSION_DATE
+#define IPVSADM_VERSION_NO		"v" VERSION
+#define IPVSADM_VERSION_DATE		"2002/11/12"
+#define IPVSADM_VERSION			IPVSADM_VERSION_NO " " IPVSADM_VERSION_DATE
 
-#define MINIMUM_IPVS_VERSION_MAJOR      0
-#define MINIMUM_IPVS_VERSION_MINOR      9
-#define MINIMUM_IPVS_VERSION_PATCH      6
+#define MINIMUM_IPVS_VERSION_MAJOR      1
+#define MINIMUM_IPVS_VERSION_MINOR      0
+#define MINIMUM_IPVS_VERSION_PATCH      4
 
 #ifndef IPVS_VERSION
 #define IPVS_VERSION(x,y,z) (((x)<<16)+((y)<<8)+(z))
@@ -173,7 +167,8 @@ static const char* cmdnames[] = {
 #define OPT_DAEMON	0x00800
 #define OPT_STATS	0x01000
 #define OPT_RATE	0x02000
-#define NUMBER_OF_OPT	14
+#define OPT_SORT	0x04000
+#define NUMBER_OF_OPT	15
 
 static const char* optnames[] = {
 	"numeric",
@@ -190,6 +185,7 @@ static const char* optnames[] = {
 	"daemon",
 	"stats",
 	"rate",
+	"sort",
 };
 
 /*
@@ -202,22 +198,22 @@ static const char* optnames[] = {
  */
 static const char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 {
-	/*     -n   -c   svc  -s   -p   -M   -r   fwd  -w   -mc  -to  dmn  -st -rt */
-/*INSERT*/    {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x'},
-/*ADD*/       {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*EDIT*/      {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*DEL*/       {'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*FLUSH*/     {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*LIST*/      {' ', '1', '1', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', '1', ' ', ' '},
-/*ADD-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x'},
-/*DEL-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*EDIT-SRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x'},
-/*TIMEOUT*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*START-D*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' ', 'x', 'x', 'x', 'x'},
-/*STOP-D*/    {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*RESTORE*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*SAVE*/      {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*ZERO*/      {'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+	/*     -n   -c   svc  -s   -p   -M   -r   fwd  -w   -mc  -to  dmn  -st  -rt  srt */
+/*INSERT*/    {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x'},
+/*ADD*/       {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*EDIT*/      {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*DEL*/       {'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*FLUSH*/     {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*LIST*/      {' ', '1', '1', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', '1', ' ', ' ', ' '},
+/*ADD-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*DEL-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*EDIT-SRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*TIMEOUT*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*START-D*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x'},
+/*STOP-D*/    {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*RESTORE*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*SAVE*/      {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*ZERO*/      {'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
 };
 
 /* printing format flags */
@@ -226,6 +222,7 @@ static const char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 #define FMT_RULE	0x0002
 #define FMT_STATS	0x0004
 #define FMT_RATE	0x0008
+#define FMT_SORT	0x0010
 
 #define SERVICE_NONE    0x0
 #define SERVICE_ADDR    0x1
@@ -251,7 +248,7 @@ static char * port_to_anyname(unsigned short port, unsigned short proto);
 static char * addrport_to_anyname(struct in_addr *addr, unsigned short port,
 				  unsigned short proto, unsigned int format);
 static int parse_service(char *buf, u_int16_t proto,
-		  u_int32_t *addr, u_int16_t *port);
+			 u_int32_t *addr, u_int16_t *port);
 static int parse_netmask(char *buf, u_int32_t *addr);
 static int parse_timeout(char *buf, int min, int max);
 static unsigned int parse_fwmark(char *buf);
@@ -357,6 +354,7 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		{"daemon", '\0', POPT_ARG_NONE, NULL, '6'},
 		{"stats", '\0', POPT_ARG_NONE, NULL, '7'},
 		{"rate", '\0', POPT_ARG_NONE, NULL, '8'},
+		{"sort", '\0', POPT_ARG_NONE, NULL, '9'},
 		{NULL, 0, 0, NULL, 0}
 	};
 
@@ -474,7 +472,7 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 				fail(2, "illegal real server "
 				     "address[:port] specified");
 			/* copy vport to dport if none specified */
-			if (parse == 1)
+			if (parse == SERVICE_ADDR)
 				ur->dport = ur->vport;
 			break;
 		case 'i':
@@ -520,6 +518,10 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			set_option(options, OPT_RATE);
 			*format |= FMT_RATE;
 			break;
+		case '9':
+			set_option(options, OPT_SORT);
+			*format |= FMT_SORT;
+			break;
 		default:
 			fail(2, "invalid option");
 		}
@@ -566,8 +568,7 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 {
 	int c, parse;
 	const char *optstring = "AEDCZSRaedlLhvt:u:f:s:M:p::w:r:gmicn";
-	const struct option long_options[] =
-	{
+	const struct option long_options[] = {
 		{"add-service", 0, 0, 'A'},
 		{"edit-service", 0, 0, 'E'},
 		{"delete-service", 0, 0, 'D'},
@@ -602,6 +603,7 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		{"daemon", 0, 0, '6'},
 		{"stats", 0, 0, '7'},
 		{"rate", 0, 0, '8'},
+		{"sort", 0, 0, '9'},
 		{"help", 0, 0, 'h'},
 		{0, 0, 0, 0}
 	};
@@ -788,6 +790,10 @@ parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			set_option(options, OPT_RATE);
 			*format |= FMT_RATE;
 			break;
+		case '9':
+			set_option(options, OPT_SORT);
+			*format |= FMT_SORT;
+			break;
 		default:
 			fail(2, "invalid option");
 		}
@@ -933,12 +939,12 @@ static int parse_timeout(char *buf, int min, int max)
 
 	/* it is just for parsing timeout of persistent service */
 	if (buf == NULL)
-		return IP_VS_TEMPLATE_TIMEOUT;
+		return IPVS_SVC_PERSISTENT_TIMEOUT;
 
 	if ((i=string_to_number(buf, min, max)) == -1)
 		fail(2, "invalid timeout value `%s' specified", buf);
 
-	return i * HZ;
+	return i;
 }
 
 
@@ -1166,6 +1172,7 @@ static void usage_exit(const char *program, const int exit_status)
 		"  --daemon                            output of daemon information\n"
 		"  --stats                             output of statistics information\n"
 		"  --rate                              output of rate information\n"
+		"  --sort                              sorting output of service/server entries\n"
 		"  --numeric      -n                   numeric output of addresses and ports\n",
 		DEF_SCHED);
 
@@ -1267,9 +1274,10 @@ static void print_conn(char *buf, unsigned int format)
 	int n;
 	unsigned long temp1, temp2, temp3;
 	char *cname, *vname, *dname;
-	unsigned long minutes, seconds, sec100s;
+	unsigned int	minutes, seconds;
+	char		expire_str[12];
 
-	if ((n = sscanf(buf, " %s %lX %hX %lX %hX %lX %hX %s %d",
+	if ((n = sscanf(buf, "%s %lX %hX %lX %hX %lX %hX %s %d",
 			protocol, &temp1, &cport, &temp2, &vport,
 			&temp3, &dport, state, &expires)) == -1)
 		exit(1);
@@ -1292,13 +1300,12 @@ static void print_conn(char *buf, unsigned int format)
 	if (!(dname=addrport_to_anyname(&daddr, dport, proto, format)))
 		exit(1);
 
-	sec100s = ((expires % HZ) * 100) / HZ;
-	seconds = (expires / HZ) % 60;
-	minutes = expires / (60 * HZ);
+	seconds = expires % 60;
+	minutes = expires / 60;
+	sprintf(expire_str, "%02d:%02d", minutes, seconds);
 
-	printf("%-3s %02ld:%02ld.%02ld %-11s %-17s %-17s %-17s\n",
-	       protocol, minutes, seconds, sec100s, state,
-	       cname, vname, dname);
+	printf("%-3s %-6s %-11s %-18s %-18s %s\n",
+	       protocol, expire_str, state, cname, vname, dname);
 
 	free(cname);
 	free(vname);
@@ -1324,7 +1331,7 @@ void list_conn(unsigned int format)
 		exit(1);
 	}
 	printf("IPVS connection entries\n");
-	printf("pro expire   %-11s %-17s %-17s %-17s\n",
+	printf("pro expire %-11s %-18s %-18s %s\n",
 	       "state", "source", "virtual", "destination");
 
 	/*
@@ -1420,6 +1427,10 @@ print_service(struct ip_vs_service_user *svc, unsigned int format)
 	if (!(d = ipvs_get_dests(svc)))
 		exit(1);
 
+	if (format & FMT_SORT) {
+		ipvs_sort_dests(d, ipvs_cmp_dests);
+	}
+
 	if (svc->fwmark) {
 		if (format & FMT_RULE)
 			sprintf(svc_name, "-f %d", svc->fwmark);
@@ -1450,7 +1461,7 @@ print_service(struct ip_vs_service_user *svc, unsigned int format)
 	if (format & FMT_RULE) {
 		printf("-A %s -s %s", svc_name, svc->sched_name);
 		if (svc->flags & IP_VS_SVC_F_PERSISTENT) {
-			printf(" -p %u", svc->timeout/HZ);
+			printf(" -p %u", svc->timeout);
 			if (svc->netmask != (unsigned long int) 0xffffffff) {
 				struct in_addr mask;
 				mask.s_addr = svc->netmask;
@@ -1474,7 +1485,7 @@ print_service(struct ip_vs_service_user *svc, unsigned int format)
 	} else {
 		printf("%s %s", svc_name, svc->sched_name);
 		if (svc->flags & IP_VS_SVC_F_PERSISTENT) {
-			printf(" persistent %u", svc->timeout/HZ);
+			printf(" persistent %u", svc->timeout);
 			if (svc->netmask != (unsigned long int) 0xffffffff) {
 				struct in_addr mask;
 				mask.s_addr = svc->netmask;
@@ -1556,6 +1567,10 @@ static void list_all(unsigned int format)
 		exit(1);
 	}
 
+	if (format & FMT_SORT) {
+		ipvs_sort_services(get, ipvs_cmp_services);
+	}
+
 	print_title(format);
 	for (i=0; i<get->num_services; i++)
 		print_service(&get->entrytable[i], format);
@@ -1570,7 +1585,7 @@ void list_timeout(void)
 	if (!(u = ipvs_get_timeouts()))
 		exit(1);
 	printf("Timeout (tcp tcpfin udp): %d %d %d\n",
-	       u->tcp_timeout/HZ, u->tcp_fin_timeout/HZ, u->udp_timeout/HZ);
+	       u->tcp_timeout, u->tcp_fin_timeout, u->udp_timeout);
 	free(u);
 }
 
